/* This C application program calls CSL to extract all the    */
/* files from the current directory                           */
/*                                                            */
/* NOTE: Link together with DMSCSL                            */
/*                                                            */
/* NOTE2: THIS LIBRARY IS MEANT TO BE INTEGARTED IN ZIP/UNZIP */
/*        FOR BETER IO ROUTINES USING CSL CALLS!              */
/*        IF SOMEBODY CONTINUES FURTHER DEVELOPMENT           */
/*        PLEASE FEEL FREE TO EXTEND THE LIBRARY.             */
/*                                                            */
/*                                                            */
/*                                     George Petrov (c)1995  */

/*************************************************/
/* Possible return codes from CSL                */
/*                                               */
/* -07    Routine not loaded                     */
/* -09    Insufficient virtual storage available */
/* -10    Too many parameters specified          */
/* -11    Not enough parameters specified        */
/* -12    CSL does not exist on the release.     */
/* -13    Invalid parameter list format          */
/*************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "c370csl.h"

#pragma linkage(DMSCSL,OS)

extern int DMSCSL(const char *RTNNAME, int *RC, ...);

DIR *opendir(const char *dirname)
{
/* DMSGETDF */
   char FILEID[17] = "", FM_NO;
   char DIRTOKEN[8] = "";
/* Share variables for DMSCSL calls */
   int RETCODE = 0, REASCODE = 0;

   DIR *dirp;
   struct dirent *new;

   printf("search dir : %s\n",dirname);
   dirp = malloc( sizeof(DIR) );
   if (dirp != NULL) {
      dirp->D_list = dirp->D_curpos = NULL;
      strcpy( dirp->D_path, dirname );
   }
   else {
       errno = ENOMEM;
       return NULL;
   }
   DMSCSL("DMSOPDIR", &RETCODE, &REASCODE, dirname, strlen(dirname),
                      "FILE", 4, dirp->D_token);

   if (RETCODE == -12) {  /* CSL calls does not exist in this release */

   }
   else if (RETCODE != 0) {
     printf ("After OPERDIR  call:\n");
     printf ("RETCODE = %d\n",RETCODE);
     printf ("REASCODE = %d\n",REASCODE);

     errno = ENOENT;
     return NULL;
   }

    if (RETCODE == 0) {
       do {
         DMSCSL("DMSGETDF", &RETCODE, &REASCODE, dirp->D_token,
                          FILEID, &FM_NO);
         if (RETCODE == 0 || (RETCODE == 4 && REASCODE == 44040)) {
             new = malloc( sizeof(struct dirent) );
             if (dirp->D_list == NULL)
                dirp->D_list = dirp->D_curpos = new;
             else
                dirp->D_curpos = (dirp->D_curpos->d_next = new);
             new->d_next = NULL;
             memcpy( new->d_name, FILEID, 8 );
             new->d_name[9] = '\0';
             strcat(new->d_name, " ");
             strncat(new->d_name, &(FILEID[8]), 8 );
             new->d_name[18] = '\0';
         }
          else {
            printf ("After GETDF call:\n");
            printf ("RETCODE = %d\n",RETCODE);
            printf ("REASCODE = %d\n",REASCODE);
          }
       } while (RETCODE == 0);
       dirp->D_curpos = dirp->D_list;
       return dirp;
    }
}

struct dirent *readdir(DIR *dirp)
{
   struct dirent *cur;

   if (!dirp) return NULL;
   cur = dirp->D_curpos;
   dirp->D_curpos = dirp->D_curpos->d_next;
   return cur;
}

void rewinddir(DIR *dirp)
{
   if (!dirp) return;
   dirp->D_curpos = dirp->D_list;
}

int closedir(DIR *dirp)
{
   struct dirent *node;
/* Share variables for DMSCSL calls */
   int RETCODE, REASCODE, BUFSIZE, DIRENT;

   if (!dirp) return 0;
   while (dirp->D_list != NULL) {
      node = dirp->D_list;
      dirp->D_list = dirp->D_list->d_next;
      free( node );
   }
   DMSCSL("DMSCLDIR", &RETCODE, &REASCODE, dirp->D_token,
                     "COMMIT", 6);

   free( dirp );
   if (RETCODE != 0) {
        printf ("After CLOSEDIR call:\n");
        printf ("RETCODE = %d\n",RETCODE);
        printf ("REASCODE = %d\n",REASCODE);
        return -1;
   }
   return 0;
}

int fstat(FILE *handle, struct stat *buf)
{
#ifndef NOCSL
/* Share variables for DMSCSL calls */
   int   RETCODE = 0;
   int   REASCODE = 0;
/* variables for DMSEXIST */
   char  cslbuf[416];
   int   lyy, lmo, ldd, lhh, lmm, lss;
#endif

long int curpos;
int rc = 0;
time_t ltime;

buf->st_dev = 0;
buf->st_ino = 0;
buf->st_uid = 0;
buf->st_gid = 0;
buf->st_rdev = buf->st_dev;
buf->st_mode = S_IFREG;
buf->st_nlink = 1;
buf->st_size = 0;

#ifdef NOCSL
if ((rc = fldata(handle,buf->fname,&(buf->attr))) != 0) {
   buf->st_size = 0;
   return rc;
}
if (buf->attr.__recfmF) {
   if ((rc = curpos = ftell(handle)) == 0) {
      rc = fseek(handle,0L,SEEK_END);
      buf->st_size = ftell(handle);
      if (rc == 0 &&
         (rc = fseek(handle,curpos,SEEK_SET)) == 0);
      }
   }
time(&ltime);
buf->st_atime =
buf->st_mtime =
buf->st_ctime = ltime;

#else /* Use CSL */

memset((void *)&(buf->attr), 0, sizeof(fldata_t));
strcpy(buf->fname, (char *)handle);

DMSCSL("DMSEXIST", &RETCODE, &REASCODE, buf->fname, strlen(buf->fname),
                 cslbuf, 416, "NOCOMMIT", 8);
if (RETCODE == 0) {
    if (cslbuf[0] == '1') { /* FILE Data Record */
       buf->attr.__dsorgPDSmem = 1;
       buf->attr.__recfmF = (cslbuf[21] == 'F' ? 1 : 0);
       buf->attr.__recfmV = (cslbuf[21] == 'V' ? 1 : 0);
       buf->attr.__maxreclen = *((int *)&cslbuf[24]);
       buf->attr.__blksize   = buf->attr.__recfmV ?
                               buf->attr.__maxreclen + 4 :
                               buf->attr.__maxreclen;

       /* most recently update */
       sscanf (cslbuf+44, "%2d/%2d/%2d%2d:%2d:%2d",
                           &lyy,&lmo,&ldd,&lhh,&lmm,&lss);
       buf->st_mtime = dostime(lyy, lmo, ldd, lhh, lmm, lss);

       /* most recently access */
       sscanf (cslbuf+323, "%2d/%2d/%2d%2d:%2d:%2d",
                           &lyy,&lmo,&ldd,&lhh,&lmm,&lss);
       buf->st_atime = dostime(lyy, lmo, ldd, lhh, lmm, lss);

       /* create date/time */
       sscanf (cslbuf+92, "%2d/%2d/%2d%2d:%2d:%2d",
                           &lyy,&lmo,&ldd,&lhh,&lmm,&lss);
       buf->st_ctime = dostime(lyy, lmo, ldd, lhh, lmm, lss);

       /* Read and Write Authority */
       if (cslbuf[69] == '1') buf->st_mode |= S_IREAD;
       if (cslbuf[70] == '1') buf->st_mode |= S_IWRITE;

       /* File size */
       buf->st_size = buf->attr.__recfmF ?
          buf->attr.__maxreclen * *((int *)&cslbuf[32]) : /* num_recs */
          (*((int *)&cslbuf[28])-1) * 4096;    /* (num_blks-1) * 4096 */
    }
    else {   /* DIRECTORY Data Record */
       buf->attr.__dsorgPDSdir = 1;
       buf->st_mode |= S_IFDIR;
       buf->st_mode &= ~S_IFREG;
       /* Read and Write Authority */
       if (cslbuf[158] == '1') buf->st_mode |= S_IREAD;
       if (cslbuf[159] == '1') buf->st_mode |= S_IWRITE;

       /* most recently update */
       /*  & most recently access */
       sscanf (cslbuf+214, "%2d/%2d/%2d%2d:%2d:%2d",
                           &lyy,&lmo,&ldd,&lhh,&lmm,&lss);
       buf->st_mtime =
       buf->st_atime = dostime(lyy, lmo, ldd, lhh, lmm, lss);

       /* create date/time */
       sscanf (cslbuf+238, "%2d/%2d/%2d%2d:%2d:%2d",
                           &lyy,&lmo,&ldd,&lhh,&lmm,&lss);
       buf->st_ctime = dostime(lyy, lmo, ldd, lhh, lmm, lss);
    }
}
#endif /* NOCSL */
return rc;
}

int stat(char *path, struct stat *buf)
{
  FILE *stream;
  int rc;
#ifndef NOCSL
  stream = (FILE *)path;
#else
  if ((stream = fopen(path, "rb")) == NULL)
  {
     return -1;
  }
#endif /* NOCSL */
  rc = fstat(stream, buf);
#ifdef NOCSL
  fclose(stream);
#endif /* NOCSL */
  return rc;
}

unsigned long dostime(y, n, d, h, m, s)
int y;                  /* year */
int n;                  /* month */
int d;                  /* day */
int h;                  /* hour */
int m;                  /* minute */
int s;                  /* second */
/* Convert the date y/n/d and time h:m:s to a four byte DOS date and
   time (date in high two bytes, time in low two bytes allowing magnitude
   comparison). */
{
  return y < 1980 ? dostime(1980, 1, 1, 0, 0, 0) :
        (((ulg)y - 1980) << 25) | ((ulg)n << 21) | ((ulg)d << 16) |
        ((ulg)h << 11) | ((ulg)m << 5) | ((ulg)s >> 1);
}

/* Change to directory 'path', access */
/* the first free disk for it and     */
/* return the disk letter in 'disk'   */
int chdir(char *path, char *disk)
{
/* Share variables for DMSCSL calls */
   int   RETCODE = 0;
   int   REASCODE = 0;
/* variables for DMSEXIST */
   char  cslbuf[416];
   char  free_disk;
/* Check if the path exist first */
/* Use DMSEXIST */

DMSCSL("DMSEXIST", &RETCODE, &REASCODE, path, strlen(path),
                 cslbuf, 416, "NOCOMMIT", 8);

/* if the path exist, access it under a free disk letter */
if (RETCODE == 0 && cslbuf[0] == '2') {
/* if the directory is already accessed - just return the disk */
   strcpy(cslbuf, "ACCESS ");
   strcat(cslbuf, path);
   strcat(cslbuf, "  ");
   if (FindFreeDisk(&free_disk) != 0) return -1;
   cslbuf[strlen(cslbuf)-1] = free_disk;
   if (system(cslbuf) != 0) return -1;
   return 0;
}
else return -1;
}

/* Get the directory accessed under disk 'disk' */
/* return the name in 'path' (first allocates enough space */
int getcurdir(char *disk, char *path)
{
/* Use DMSQFMOD */
}

/* Creates directory specified in 'path' */
int mkdir(char *path)
{
   int   moreParent = 0, len;
   char  *next, *end;
/* Share variables for DMSCSL calls */
   int   RETCODE = 0;
   int   REASCODE = 0;

   end = path + strlen(path);
   next = strchr(path, '.');
   next = strchr(next+1, '.');
   do {
       if (next == NULL) next = end;
       DMSCSL("DMSCRDIR", &RETCODE, &REASCODE,
                    path, next-path, "COMMIT", 6);
       if (next != end ) next = strchr(next+1,'.');
   } while (next != end && (RETCODE == 0 || REASCODE == 20000));
   if (RETCODE != 0) {
      if (REASCODE == 20000) return 0;    /* dir already exist */
      else return -1;
   }
   else return 0;
}

/* returns the first free disk letter in 'disk' */
int FindFreeDisk(char *disk)
{
/* Share variables for DMSCSL calls */
   int   RETCODE = 0;
   int   REASCODE = 0;
/* DMSQFMOD */
   char letters[]   = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
   char cslbuf[153] = "";
   int  disk_flag, i;

   for (i=0; i<26; i++) {
        DMSCSL("DMSQFMOD", &RETCODE, &REASCODE,
                letters+i, cslbuf, &disk_flag);
        if (RETCODE != 0) {
           if (REASCODE == 90700) {
              *disk = letters[i];
              return 0;
           }
           else {
              return -1;
           }
        }
   }
   return -1;
}

char *dir2longdir(char *dir)
{
/* Share variables for DMSCSL calls */
   int   RETCODE = 0;
   int   REASCODE = 0;
/* DMSERP   */
   char * newdir;
   char  userid[9];
   char  buf[1024];
   int   buflen = 1024;

   newdir = (char *)malloc(153);
   strncpy(userid, (char *)0x5ED8, 8);
   userid[9] = '\0';

   if (strchr(dir, ':') == NULL) {
      DMSCSL("DMSERP  ", &RETCODE, "RESET   ");
      DMSCSL("DMSERP  ", &RETCODE, "EXTRACT ", 1,
             "CURRENT_FILE_POOL_ID", buf, 32, &buflen);
      if (RETCODE != 0) {
           printf ("After RESET call:\n");
           printf ("RETCODE = %d\n",RETCODE);
           printf ("REASCODE = %d\n",REASCODE);
           return NULL;
      }
      buf[9] = '\0';
      strcpy(newdir,buf);
      if (dir[0] == '.')
         strcat(newdir,userid);

      printf("long dir : %s\n", newdir);
   }
}
