/*  C370CSL.H

    Definitions used for file status functions

*/

#ifndef __C370CSL_H
#define __C370CSL_H

#include <stdio.h>
#include <time.h>

#define ulg unsigned long

/* <stat.h> definitions */

#define S_IFMT  0xF000  /* file type mask */
#define S_IFDIR 0x4000  /* directory */
#define S_IFIFO 0x1000  /* FIFO special */
#define S_IFCHR 0x2000  /* character special */
#define S_IFBLK 0x3000  /* block special */
#define S_IFREG 0x8000  /* or just 0x0000, regular */
#define S_IREAD 0x0100  /* owner may read */
#define S_IWRITE 0x0080 /* owner may write */
#define S_IEXEC 0x0040  /* owner may execute <directory search> */

struct  stat
{
    short st_dev;      /* Drive number of disk containing the  */
                       /* file or file handle if the file is   */
                       /* on device                            */
    short st_ino;      /* File number - 0..6                   */
    short st_mode;     /* Bit mask giving information about    */
                       /* the file's mode                      */
    short st_nlink;    /* Set to the integer constant 1        */
    int   st_uid;      /* Owner of the file                    */
    int   st_gid;      /* same as st_uid                       */
    short st_rdev;     /* Same as st_dev                       */
    long  st_size;     /* Size of the file in bytes            */
    long  st_atime;    /* Most recent access                   */
    long  st_mtime;    /* Last modifyed time                   */
    long  st_ctime;    /* Create time                          */
    FILE   *fp;
    char   fname[FILENAME_MAX];  /* path and filename together */
    fldata_t attr;     /* All attributes for this file         */
};

int  fstat(FILE *handle, struct stat *buf);
int  stat(char *path, struct stat *buf);
unsigned long dostime(int y, int n, int d, int h, int m, int s);
int chdir(char *path, char *disk);
int getcurdir(char *disk, char *path);
int mkdir(char *path);
int FindFreeDisk(char *disk);
char *dir2longdir(char *dir);

/* <dirent.h> definitions */

#define NAMELEN     16

struct dirent {
   struct dirent *d_next;
   char   d_name[NAMELEN+1];
};

typedef struct _DIR {
   struct  dirent *D_list;          /* the begining of the dir lst */
   struct  dirent *D_curpos;             /* current file pointer   */
   char            D_path[FILENAME_MAX]; /* the whole path         */
   char            D_token[8];           /* used as SFS dir token  */
} DIR;

DIR *          opendir(const char *dirname);
struct dirent *readdir(DIR *dirp);
void           rewinddir(DIR *dirp);
int            closedir(DIR *dirp);

/* fcntl.h     definitions   */

#define  O_RDONLY   0x0001   /* Open file for Read Only       */
#define  O_WRONLY   0x0002   /* Open file for Write Only      */
#define  O_RDWR     0x0004   /* Open file for Read and Write  */

#define  O_CREAT    0x0100   /* Create new file               */
#define  O_TRUNC    0x0200
#define  O_EXCL     0x0400   /* Open file for exclusively use */
#define  O_APPEND   0x0800   /* Append to the file            */

#define  O_BINARY   0x1000   /* Open file as binary           */
#define  O_MEMORY   0x2000   /* Open a memory file            */
#define  O_RECORD   0x4000   /* Open file for record i/o      */
#define  O_DCB      0x8000   /* DCB parameters are used       */

extern unsigned char dcb_flags[];       /* used by fopen and O_DCB */
   /* with a leading comma, eg:  ",recfm=fb,lrecl=80,blksize=6160" */

/* io.h        definitions   */

#define  fileno(fp)   ((int)fp)
#define  isatty(fd)   (fd==((int)stdin) || fd==((int)stdout) || fd==((int)stderr
#define  chmod(a,b)   0
#define  close(fd)    fclose((FILE *)fd)
#define  eof(fd)      feof((FILE *)fd)
#define  tell(fd)     ftell((FILE *)fd)
#define  unlink(fn)   remove(fn)
#define  lseek(a,b,c) fseek((FILE *)a,b,c)
#define  read(a,b,c)  fread(b,1,c,(FILE *)a)
#define  write(a,b,c) fwrite(b,1,c,(FILE *)a)
int  open(const char *path, int access);
int  setmode(int fd, int access);

/* stat.h      definitions   */

#define _FLDATA(m)   (*(fldata_t *) &m)
#define S_ISDIR(m)   (_FLDATA(m).__dsorgPDSdir)
#define S_ISREG(m)   (_FLDATA(m).__dsorgPO | \
                      _FLDATA(m).__dsorgPDSmem | \
                      _FLDATA(m).__dsorgPS)
#define S_ISBLK(m)   (_FLDATA(m).__recfmBlk)
#define S_ISMEM(m)   (_FLDATA(m).__dsorgMem)

/* errno.h     definitions   */

#define EZERO    0      /* Error 0                  */
#define EINVFNC  1      /* Invalid function number  */
#define ENOFILE  2      /* File not found           */
#define ENOPATH  3      /* Path not found           */
#define ECONTR   7      /* Memory blocks destroyed  */
#define EINVMEM  9      /* Invalid memory block address */
#define EINVENV 10      /* Invalid environment      */
#define EINVFMT 11      /* Invalid format           */
#define EINVACC 12      /* Invalid access code      */
#define EINVDAT 13      /* Invalid data             */
#define EINVDRV 15      /* Invalid drive specified  */
#define ECURDIR 16      /* Attempt to remove CurDir */
#define ENOTSAM 17      /* Not same device          */
#define ENMFILE 18      /* No more files            */

#define ENOENT   2      /* No such file or directory*/
#define EMFILE   4      /* Too many open files      */
#define EACCES   5      /* Permission denied        */
#define EBADF    6      /* Bad file number          */
#define ENOMEM   8      /* Not enough core          */
#define EFAULT  14      /* Unknown error            */
#define ENODEV  15      /* No such device           */
#define EINVAL  19      /* Invalid argument         */
#define E2BIG   20      /* Arg list too long        */
#define ENOEXEC 21      /* Exec format error        */
#define EXDEV   22      /* Cross-device link        */
#define ENFILE  23      /* Too many open files      */
#define ECHILD  24      /* No child process         */
#define ENOTTY  25      /* UNIX - not MSDOS         */
#define ETXTBSY 26      /* UNIX - not MSDOS         */
#define EFBIG   27      /* UNIX - not MSDOS         */
#define ENOSPC  28      /* No space left on device  */
#define ESPIPE  29      /* Illegal seek             */
#define EROFS   30      /* Read-only file system    */
#define EMLINK  31      /* UNIX - not MSDOS         */
#define EPIPE   32      /* Broken pipe              */
#define EEXIST  35      /* File already exists      */
#define EDEADLOCK 36    /* Locking violation        */
#define EPERM   37      /* Operation not permitted  */
#define ESRCH   38      /* UNIX - not MSDOS         */
#define EINTR   39      /* Interrupted function call */
#define EIO     40      /* Input/output error       */
#define ENXIO   41      /* No such device or address */
#define EAGAIN  42      /* Resource temporarily unavailable */
#define ENOTBLK 43      /* UNIX - not MSDOS         */
#define EBUSY   44      /* Resource busy            */
#define ENOTDIR 45      /* UNIX - not MSDOS         */
#define EISDIR  46      /* UNIX - not MSDOS         */
#define EUCLEAN 47      /* UNIX - not MSDOS         */

#endif /* C370CSL_H */
